/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_IOSOCKSTREAM_HPP
#define OAPR_IOSOCKSTREAM_HPP
#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include <iostream>
#include "boost/scoped_ptr.hpp"
#include "oapr_exception.hpp"
#include "oapr_exc_utils.hpp"
#include "oapr_sockbuf.hpp"
FILE_VERSION("$Revision: 1.4 $",oapr_iosockstream_hpp);

namespace oapr
{

  template<typename charT,typename traits,
	   typename thread_pol_t=thread_pol_none,
	   std::streamsize buf_sz=255>
  class basic_iosockstream : public std::basic_iostream<charT,traits>
  {
  public:

    //standard typedefs
    typedef charT                     char_type;
    typedef typename traits::int_type int_type;
    typedef typename traits::pos_type pos_type;
    typedef typename traits::off_type off_type;
    typedef traits                    traits_type;
    typedef thread_pol_t              thread_pol_type;

    //non standard typdefs
    //typedef put stream buf type here
    typedef std::basic_ios<char_type,traits>            ios_type;
    typedef std::basic_iostream<char_type,traits>       iostream_type;
    typedef std::istreambuf_iterator<char_type,traits>  istreambufiter_type;
    typedef std::ostreambuf_iterator<char_type,traits>  ostreambufiter_type;
    typedef basic_sockbuf<char_type,traits_type,
			  thread_pol_type,buf_sz>       sockbuf_type;
    typedef basic_iosockstream<char_type,traits_type,
			       thread_pol_type,buf_sz>  iosockstream_type;

    typedef pool_mgr::pool_t pool_t;
    typedef pool_mgr::pool_ptr pool_ptr;
  private:
    sockinfo m_si;
    boost::scoped_ptr<sockbuf_type> m_sockbuf;
  public:
    basic_iosockstream(pool_ptr& p)
      : iostream_type(NULL)
	,m_si(p)
	,m_sockbuf(new sockbuf_type(p)){
      this->init(m_sockbuf.get());
    } 
    basic_iosockstream(pool_ptr& p,const char* host,unsigned short port,
		       int type, typename sockbuf_type::sockmodes md)
      : iostream_type(NULL)
	,m_si(p)
	,m_sockbuf(new sockbuf_type(p)){
      try{
	this->init(m_sockbuf.get());
	this->open(host,port,type,md);
      }OAPR_CATCH();
    }
	

    virtual ~basic_iosockstream(){}
    virtual sockbuf_type* rdbuf()const{
      return const_cast<sockbuf_type*>(m_sockbuf.get());
    }

    virtual void open(const char* host,
		      unsigned short port,
		      int type, typename sockbuf_type::sockmodes md);
    virtual bool is_open()const{
      return m_sockbuf->is_created()&&m_sockbuf->is_open();
    }
    virtual int option(int opt, int on);
    virtual int close();
    virtual int shutdown(apr_shutdown_how_e how);
    virtual int create(int fam, int type);
    virtual int connect();

    //specific to in
    virtual int listen(int backlog);
    virtual int bind();
    virtual iosockstream_type* accept();

  protected:
    basic_iosockstream(sockbuf_type* sb)
      : iostream_type(NULL)
	,m_si(sb->gpool())
	,m_sockbuf(sb){
      this->init(m_sockbuf.get());
    }

  };

  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> void
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::open(const char* host,
					 unsigned short port,
					 int type, 
					 typename sockbuf_type::sockmodes md)
  {
    /* verifying that the socket is not open already*/
    if(is_open()){
      this->setstate(std::ios_base::failbit);
      return;
    }
    /* create a sockinfo and use that to create a socket*/
    m_si.create(host,APR_UNSPEC,port,0);
    /**@todo put exception catch and status check */
    OAPR_CHECK(this->create(m_si.family(),type));

    if(md==sockbuf_type::client){
      OAPR_CHECK(this->connect());
    }else if(md==sockbuf_type::server){
      OAPR_CHECK(this->bind());
      OAPR_CHECK(this->listen(5));
    }else if(md==sockbuf_type::none){
    }else{
      this->setstate(std::ios_base::failbit);
    }
  }


  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::option(int opt, int on)
  {
    try{
      return m_sockbuf->option(opt,on);
    }OAPR_CATCH();
  }
  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::close(){
    try{
      return m_sockbuf->close();
    }OAPR_CATCH();
  }
  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::shutdown(apr_shutdown_how_e how){
    try{
      return m_sockbuf->shutdown(how);
    }OAPR_CATCH();
  }
  
  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::create(int fam, int type){
    try{
      return m_sockbuf->create(fam,type);
    }OAPR_CATCH();
  }
  
  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::connect(){
    try{
      return m_sockbuf->connect(m_si);
    }OAPR_CATCH();
  }

  template<typename charT,typename traits, 
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::listen(int backlog){
    try{
      return m_sockbuf->listen(backlog);
    }OAPR_CATCH();
  }

  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz> int
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::bind(){
    try{
      return m_sockbuf->bind(m_si);
    }OAPR_CATCH();
  }

  template<typename charT,typename traits,
	   typename thread_pol_t,std::streamsize buf_sz>
  typename basic_iosockstream<charT,traits,
			      thread_pol_t,buf_sz>::iosockstream_type* 
  basic_iosockstream<charT,traits,thread_pol_t,buf_sz>::accept(){
    try{
      sockbuf_type* sb=m_sockbuf->accept();
      return new iosockstream_type(sb);
    }OAPR_CATCH();
  }
};
#endif
