/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_THREAD_HPP
#define OAPR_THREAD_HPP

#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include "oapr/oapr_pool.hpp"
#include "oapr/oapr_cond.hpp"
#include "oapr/oapr_mutex.hpp"
#include "oapr/oapr_exc_base.hpp"
#include "oapr/oapr_exc_logic.hpp"
#include "apr_thread_proc.h"
#include <iostream>
FILE_VERSION("$Revision: 1.5.2.2 $",oapr_thread_hpp);
// temp

/*
 * Two thread termination models. Override termination function for second.
 * 1: Easy: Infinite work is done in a while(!terminated) loop which periodically checks for termination state, handled by thread class. SUPPORTED
 * 2: Involved: Thread is signalled via thread method override and takes it's own action.
 */

namespace oapr
{
  enum thread_state
  {
    tsSuspended,
    tsRunning,
    tsTerminating,
    tsTerminated
  };

  /**\todo Pick a sensible name for this thread lifecycle implementation class */
  class LIBOAPR_API thread_class
  {
#ifdef WIN32
  public:
	  thread_class(){}
#endif
    public:
      typedef pool::pool_ptr pool_ptr;
    public:
      /**
       * Create a thread object. Override the execute() method with the code to run in the new thread.
       * @param start_now If true the thread is started immediately (default).
       */
      thread_class(const pool_ptr& p);
      /**
       * Destroy the thread object. Terminates the thread then frees it's associated memory.
       */
      virtual ~thread_class();
      /**
       * Start execution of a thread created with start_now = false.
       * If the thread is already running, returns with no action.
       * If the thread has been terminated, throws an
       */
      void start();
      /**
       * Tells the thread to terminate cleanly by setting terminated to
       * false, then joins the thread and waits until it has finished.
       * @retval Return code of the thread.
       */
      apr_status_t terminate();
      /**
       * Tells the thread to terminate cleanly by setting terminated to
       * false. The method returns immediately. If you require a return
       * value from the thread, or to know it has fully terminated
       * before continuing, call terminate() instead.
       */
      void signal_terminate();
      /**
        * Block the current thread until this thread stops executing and returns the exit code
       * of the thread. If the thread has already finished, returns immediately with the exit code.
       * @retval Exit code returned by the thread.
       */
      apr_status_t join();
      /**
       * Returns true if the thread is active (running), including when it has been
       * signalled to terminate and has not yet done so.
       */
      bool active();
      /**
       * @retval Returns true if the thread has been terminated, presumably through a call
       * to terminate() or delete. Also returns true if the thread has been
       * signalled to terminate and is still busy shutting down.
       * The usual way to check for a terminate signal in oapr threads is with a while(!terminated()) loop.
       * @remark Check this periodically from the thread and exit when it is true.
       * @remark Note that it's possible for active() and terminated() to both
       * return true, when termination has been requested but has not yet occurred.
       * If you need to know that a thread has fully terminated, test !active() instead.
       */
      bool terminated();
      /**
       * Returns true if thread execution is suspended (because it hasn't been started yet).
       * @retval Returns true if the thread is suspended.
       * @remark Threads are suspended when they have been constructed with
       * the start_now parameter set to false, until start() is called.
       */
      bool suspended();
      /**
       * Returns the state of the thread object. State is one of:
       * <PRE>
       *
       *        tsSuspended    --  Thread is suspended, i.e. awaiting call to start()
       *        tsRunning      --  Thread is running
       *        tsTerminating  --  Thread has been signalled to terminate but has not yet done so
       *        tsTerminated   --  Thread has finished execution and is no longer running
       * </PRE>
       * @tip Consider a call to one of active() or terminated() rather than checking the state, as the
       * state set is private to the thread class and may be extended in the future. Inside the thread
       * execute() method, state will never be tsSuspended or tsTerminated.
       */
      thread_state state();
    protected:
      /**
       * Called by thread runners to initialize the thread, starting execution.
       */
      void initialize();
      /**
       * Called by thread runners prior to destroying the thread, to be sure the thread is finished
       * and all virtual methods are finished with.
       */
      void destroy();
      /**
       * Main body of execution for a thread created with start_now = false.
       * @return apr_status_t Result of execution, passed back to anyone who cares.
       */
      virtual apr_status_t execute()=0;
      /**
       * Routine called to tell thread to terminate.
       * @retval The return code of the thread. A successful thread should return APR_SUCCESS.
       */
      virtual apr_status_t do_terminate();
      /**
       * Called on unhandled exception.
       * Override this method if you want to handle uncaught exceptions in the thread.
       * The default behaviour is to return APR_EUNCAUGHT_EXCEPTION.
       * @param base A pointer to the exception that was caught. NULL if the exception
       * is not derived from oapr::exception_base (i.e. catch(...))
       * @retval Return code of the thread.
       * @warning This method is called from the thread, and the thread object is locked.
       * Attempts to change the thread in any way will block.
       */
      virtual apr_status_t unhandled_exception(exception_base *base) throw();
      /**
       * Exit the thread.
       * @param status The status code to be returned to anyone who cares.
       */
      void exit(apr_status_t status);
    private:
      /**
       * Thread entry point into the object. Called by thread_entry_fn.
       */
      void* thread_main(apr_thread_t *thd) throw();
      /**
       * Thread entry point.
       */
      static void* APR_THREAD_FUNC thread_entry_fn(apr_thread_t *thd, void *data);

      /**
       * Thread cleanup.
       */
      void cleanup(apr_status_t exitcode) throw();
      void cleanup_exception(exception_base *e) throw();
      
      /**
       * Terminate and join apr thread, called from destructor.
       */
      void terminate_and_join_thread() throw();

      /** Main thread state. */
      thread_state  m_state;
      mutex         m_state_lock; // cannot be a rwlock, needs to be locked during m_terminated prereq check.
      /** Pointer to apr thread structure */
      apr_thread_t *m_thread;
      /**\todo Exit code */
      apr_status_t  m_exitcode;
      /** Terminated condition for joined threads */
      condition     m_terminated;
      /** Shared ptr to the pool we use */
      pool_ptr      m_pool;
  };

  template<class T> class thread: public T
  {
    public:
      /**
       * Create a new thread of execution.
       * @param start_now If true the thread is started immediately (default).
       */
      thread(const thread_class::pool_ptr& p, bool start_now = true): T(p) { initialize(); if(start_now) start(); }
      ~thread() { destroy(); }
/*      thread(bool start_now = true) { m_thread.initialize(); if(start_now) start(); }
      ~thread() { thd.destroy(); }
    private:
      T m_thread;*/
  };

  /**
   * Force the current thread to yield the processor.
   * @tip Yield sucks. Wait for some signals or socket activity.
   */
  void yield();

};

#endif
