/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_TIME_HPP
#define OAPR_TIME_HPP
#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include "apr_time.h"
#include "oapr/oapr_exception.hpp"
#include "oapr/oapr_pool.hpp"
#include <string>
FILE_VERSION("$Revision: 1.5.2.1 $",oapr_time_hpp);

namespace oapr
{
  typedef apr_interval_time_t interval;

  /**
   * Simple time object.
   * @remark This is just about the only object in oapr/apr that doesn't use pools. Its way too lightweight to be considered.
   * It will do so if/when all C++ objects are themselves allocated from pools. -MC
   */
  class time
  {
  public:
    enum initial_time_t { now };
  public:
    /** Create an empty time object */
    time()
      : m_exp(0) {}
    /** Copy a time object */
    time(const time& t)
      : m_time(t.m_time) { m_exp = new apr_time_exp_t; memcpy(m_exp, t.m_exp, sizeof(apr_time_exp_t)); }
    /**
     * Create an initialized time object
     * Time may be 
     * @param init 'now' is the only value currently allowed.
     */
    time(initial_time_t init)
      : m_exp(0) { m_time = apr_time_now(); }
    /**
     * Create an initialized time object
     * @param ansi ANSI time_t structure to initialize from
     */
    time(time_t ansi)
      : m_exp(0) { OAPR_CHECK(apr_time_ansi_put(&m_time, ansi)); }
    /** Free the time object */
    ~time() { if(m_exp) delete m_exp; }
    /** Assignment operator */
    time& operator=(const time& t)  {
      if(this != &t/*self-copy*/) {
        m_time = t.m_time;
        if(m_exp) delete m_exp;
        if(t.m_exp) {
          m_exp = new apr_time_exp_t;
          memcpy(m_exp, t.m_exp, sizeof(apr_time_exp_t));
        }
        else
          m_exp = 0;
        return *this;
      }
    }
    
    apr_int32_t usec()   { expand(); return m_exp->tm_usec; }
    apr_int32_t sec()    { expand(); return m_exp->tm_sec; }
    apr_int32_t min()    { expand(); return m_exp->tm_min; }
    apr_int32_t hour()   { expand(); return m_exp->tm_hour; }
    apr_int32_t day()    { expand(); return m_exp->tm_mday; }
    apr_int32_t month()  { expand(); return m_exp->tm_mon; }
    apr_int32_t year()   { expand(); return m_exp->tm_year; }
    apr_int32_t wday()   { expand(); return m_exp->tm_wday; }
    apr_int32_t yday()   { expand(); return m_exp->tm_yday; }
    bool is_dst()        { expand(); return m_exp->tm_isdst; }
    apr_int32_t offset() { expand(); return m_exp->tm_gmtoff; }
    void assign(time_t ansi) { OAPR_CHECK(apr_time_ansi_put(&m_time, ansi)); update_exp(); }
    inline void expand() { if(!m_exp) {m_exp = new apr_time_exp_t;} apr_time_exp_gmt(m_exp, m_time); }
    inline void compact() { if(m_exp) { delete m_exp; m_exp = NULL; } }
    /** Formats dates in the RFC822 format in an efficient manner */
    /**\todo Check if its possible to alloc a new string and write to its internal buffer */
    std::string rfc822() const { char buf[APR_RFC822_DATE_LEN]; OAPR_CHECK(apr_rfc822_date(buf, m_time)); return std::string(buf); }
    /**
     * Formats dates in the ctime() format in an efficient manner.
     * Unlike ANSI/ISO C ctime(), apr_ctime() does not include a \n at the end of the string.
     */
    /**\todo Check if its possible to alloc a new string and write to its internal buffer */
    std::string ctime() const { char buf[APR_CTIME_LEN]; OAPR_CHECK(apr_ctime(buf, m_time)); return std::string(buf); }
    /**
     * Same as ctime()
     */
    inline std::string str() const { return ctime(); }
    /**
     * Formats the time according to the format specified.
     * @param buf string buffer to write to
     * @param size sizeof(buf) (The maximum length of the string)
     * @param format The format for the time string
     * @retval The length of the returned string
     */
    /**\todo Check in apr code to make sure the "the maximum length of the string" includes the \0 */
    apr_size_t strftime(char *buf, apr_size_t size, const char *format)
    {
      expand();
      apr_size_t retsize;
      OAPR_CHECK(apr_strftime(buf, &retsize, size, format, m_exp));
      return retsize;
    }
  private:
    /** Update m_exp if it is assigned (called after the time value has been changed) */
    inline void update_exp() { if(m_exp) apr_time_exp_gmt(m_exp, m_time); }
    /** 64-bit apr time structure */
    apr_time_t m_time;
    /** Expanded time structure - allocated when used or after explicit expand() call. Freed if compact() is called. */
    apr_time_exp_t *m_exp;
  };
  
  /**
   * Sleep for the specified number of micro-seconds.
   * @param t desired amount of time to sleep.
   * @warning May sleep for longer than the specified time.
   */
  inline void sleep(interval t) { apr_sleep(t); }
  /**
   * rfc822_date formats dates in the RFC822 format in an efficient manner.
   * @param t the date/time structure to convert
   * @return formatted date/time
   */
  inline std::string rfc822_date(const time& t) { return t.rfc822(); }
  /**
   * ctime formats dates in the ctime() format in an efficient manner.
   * Unlike ANSI/ISO C ctime(), apr_ctime() does not include a \n at the end of the string.
   * @param t the date/time structure to convert
   * @return formatted date/time
   */
  inline std::string ctime(const time& t) { return t.ctime(); }
  
  /**
   * Improve the clock resolution for the lifetime of the given pool.
   * Generally this is only desireable on benchmarking and other very
   * time-sensitive applications, and has no impact on most platforms.
   * @param p The pool to associate the finer clock resolution 
   */
  inline void time_clock_hires(pool& p) { apr_time_clock_hires(p.getp()); }

};
#endif
