/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_TAGLIST_HPP
#define OAPR_TAGLIST_HPP

#include <string>
#include "boost/shared_ptr.hpp"

namespace oapr
{
  /**
   * Intended uses of tag lists:
   *   Static default configuration set
   *   Configuration mapping structure to global variables
   */

  // Pre-defined tag value types
  enum _tag_types {
    tt_null    = 0x00000000,
    tt_int     = 0x00000001,  // Integer
    tt_pchar   = 0x00000002,  // Character array pointer
    tt_string  = 0x00000003,  // (Pointer to an owned) std::string
    tt_taglist = 0x00000004,  // Pointer to a const tag list
    tt_pint    = 0x00000009,  // Pointer to an Integer
    tt_pstring = 0x0000000B,  // Pointer to a std::string

    tag_ptr_bit = 0x80000000, // Bit set when tag value is a pointer to the value rather than an inline value
    tag_type_mask = 0xF0000000
  };
  typedef unsigned int tag_type_t;

  enum builtin_tag_type_mask {
    tag_mask_int     = 0x00000001,
    tag_mask_pchar   = 0x00000002,
    tag_mask_string  = 0x00000004,
    tag_mask_taglist = 0x00000008,
    tag_mask_pint    = 0x00000010,
    tag_mask_pstring = 0x00000020
//    tag_mask_ptr_bit = 0x80000000 // Bit set when tag value is a pointer to the value rather than an inline value
  };

  typedef unsigned int tag_id_t;

  struct tag_t; // forward

  typedef union {
    int int_value;
    const char *char_ptr_value;
    std::string *string_value;
    tag_t *taglist_value;
    int *int_ptr_value;
    void *extended_ptr_value;
  } tag_value_t;

  inline void copy_tag(tag_value_t& dest, const tag_value_t& src)
  {
    // use memcpy if this ever goes over 4 bytes
    assert(sizeof(tag_value_t) == 4);
    dest.extended_ptr_value = src.extended_ptr_value;
  }

  struct tag_t
  {
    tag_id_t id;
    tag_type_t type;
    tag_value_t value;
  };
  typedef struct tag tag;

  class tag_value
  {
  public:
    tag_value(tag_value_t &value) { copy_tag(m_value, value); }
    ~tag_value() {}
  private:
    tag_value_t m_value;
  };

  struct tag_name_index
  {
    const char *name;
    tag_t *tag;
  };

  /**
   * Custom tag type class - ??? copying around = bad
   */
  class tag_type
  {
  public:
    builtin_tag_type_mask supports_cast_to_types();
    int as_int();
    std::string as_string(tag_t *tag);
  };

  /**
   * Interface for generic tagged variant value list container that could hold configuration values,
   */

  class tag_container
  {
  public:
    tag_container() {}
    virtual ~tag_container() {}
    /**
     * Get a tag value. Throws an exception if the tag doesn't exist.
     */
    virtual void get(tag_id_t id, int& value)=0;
    virtual void get(tag_id_t id, std::string& value)=0;
    virtual void get(tag_id_t id, const char *& value)=0;
    /**
     * Find a tag value.
     * @retval Returns true if the tag is found, false if the tag is not found.
     */
    virtual bool find(tag_id_t id, int& value)=0;
    virtual bool find(tag_id_t id, std::string& value)=0;
    virtual bool find(tag_id_t id, const char *& value)=0;
  };

  class rw_tag_container: public tag_container
  {
  public:
    rw_tag_container() {}
    virtual ~rw_tag_container() {}
    /**
     * Set the value of a tag.
     * @tip The tag doesn't have to exist to set it's value.
     */
    virtual void set(tag_id_t id, int value)=0;
    virtual void set(tag_id_t id, const std::string& value)=0;
    virtual void set(tag_id_t id, const char *value)=0;
    /**
     * Merge into from another container into this one
     */
    virtual void assimilate(const tag_container& from, bool override = true)=0;
    /**
     * Merge this container with another
     * @param to container to merge all tags into
     * @param override override all tags that already exist (default true)
     */
    void merge(rw_tag_container& to, bool override = true)
    {
      to.assimilate(*this, override);
    }
    /**
     * Assign the contains of another container to this one.
     */
    virtual void assign(const tag_container &from)=0;
    /** Returns true if the thing is read-only*/
    virtual bool readonly()=0;
  };

  class named_tag_container: public tag_container
  {
  public:
//    named_tag_container() {}
//    ~named_tag_container() {}
    virtual void get(const std::string& name, int& value)=0;
    virtual void get(const std::string& name, std::string& value)=0;
    virtual void get(const std::string& name, const char *& value)=0;
    virtual bool find(const std::string& name, int& value)=0;
    virtual bool find(const std::string& name, std::string& value)=0;
    virtual void find(const std::string& name, const char *& value)=0;
  };

  class named_rw_tag_container: public rw_tag_container, public named_tag_container
  {
  public:
    /**
     * Set the value of a tag
     */
    virtual void set(tag_id_t id, const std::string& name, int value)=0;
    virtual void set(tag_id_t id, const std::string& name, const std::string& value)=0;
    virtual void set(tag_id_t id, const std::string& name, const char *value)=0;
  };

  /**
   * A tag list is a one-dimensional list of expected values, such as configuration settings.
   * Each value is identified by a unique integer number.
   */
  class taglist: public tag_container
  {
    public:
      taglist(const tag static_tags[]);
      ~taglist();
      /**
       * Get a tag value. Throws an exception if the tag doesn't exist.
       */
      void get(tag_id_t id, int& value);
      void get(tag_id_t id, std::string& value);
      void get(tag_id_t id, const char *& value);
      /**
       * Find a tag value.
       * @retval Returns true if the tag is found, false if the tag is not found.
       */
      bool find(tag_id_t id, int& value);
      bool find(tag_id_t id, std::string& value);
      bool find(tag_id_t id, const char *& value);
    private:
      const tag *m_tags;
  };
};

#endif
